/*
 * This file is part of the openKropki project.
 *
 * Copyright (C) 2014, 2015, Mateusz Viste
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>   /* printf(), puts() */
#include <stdlib.h>  /* rand(), srand() */
#include <time.h>    /* time() */

#include "ai.h"  /* include self for control */


static int distance(int x1, int y1, int x2, int y2) {
  int xdist, ydist;
  if (x1 > x2) {
    xdist = x1 - x2;
  } else {
    xdist = x2 - x1;
  }
  if (y1 > y2) {
    ydist = y1 - y2;
  } else {
    ydist = y2 - y1;
  }
  return(xdist + ydist);
}


static void ai_idiot(int ai_player, struct kropki_game *game) {
  int x, y, attempts, found = 0;
  for (attempts = 0;; attempts++) {
    x = rand() % 5;
    y = rand() % 5;
    x = game->lastmovex - 2 + x;
    y = game->lastmovey - 2 + y;
    if ((x < 0) || (y < 0)) continue;
    if (x >= game->field_width) x = game->field_width - 1;
    if (y >= game->field_height) y = game->field_height - 1;
    if (kropki_getpos(game, x, y) == kropki_EMPTY) {
      found = 1;
      break;
    }
    if (attempts >= 100) { /* after 100 tries, stop */
      break;
    }
  }
  if (found == 0) {
    /* if unable to find anything randomnly, scan the field sequentially */
    if (kropki_findfirstfreefield(game, &x, &y) == 0) found = 1;
  }
  /* if a possible play move is found, do it */
  if (found != 0) {
      /* printf("AI PLAYS [%d,%d]\n", x, y); */
      kropki_setpos(game, x, y, ai_player);
    } else {
      puts("NOTHING FOUND!!!!!");
  }
}


static void ai_smart(int ai_player, struct kropki_game *game) {
  int bestscore = 0, bestmovex = -1, bestmovey = -1;
  int worstscore = 0, worstmovex = -1, worstmovey = -1;
  int x, y, z, points[2], deltascore, orgdelta;
  struct kropki_game *simgame;
  kropki_getpoints(game, points);
  orgdelta = points[ai_player - 1] - points[(3 - ai_player) - 1];
  for (y = 0; y < game->field_height; y++) {
    for (x = 0; x < game->field_width; x++) {
      /* if field not empty -> skip it */
      if (kropki_getpos(game, x, y) != kropki_EMPTY) continue;
      /* test both cases: AI plays here, as well as what if opponent would play here */
      for (z = 0; z < 2; z++) {
        simgame = kropki_copygame(game);
        if (z == 0) { /* simulate playing as AI */
          kropki_setposforce(simgame, x, y, ai_player);
        } else { /* simulate playing as opponent */
          kropki_setposforce(simgame, x, y, 3 - ai_player);
        }
        kropki_getpoints(simgame, points);
        deltascore = points[ai_player - 1] - points[(3 - ai_player) - 1];
        deltascore -= orgdelta;
        /* if z == 0, then check what would happen if AI plays here */
        if (z == 0) {
          if ((deltascore > bestscore) || (bestmovex < 0)) {
            printf("best move for ai %d found at [%d,%d] -> %d pts\n", ai_player, x, y, deltascore);
            bestmovex = x;
            bestmovey = y;
            bestscore = deltascore;
          } else if (deltascore == bestscore) { /* if same score, see if closer */
            if (distance(x, y, game->lastmovex, game->lastmovey) < distance(bestmovex, bestmovey, game->lastmovex, game->lastmovey)) {
              printf("[%d,%d] is closer to [%d,%d]\n", x, y, game->lastmovex, game->lastmovey);
              bestmovex = x;
              bestmovey = y;
            }
          }
        } else { /* z == 1, so check what would happen if opponent moves here */
          if ((deltascore < worstscore) || (worstmovex < 0)) {
            printf("potential loss found if opponent plays at [%d,%d] -> %d pts\n", x, y, deltascore);
            worstmovex = x;
            worstmovey = y;
            worstscore = deltascore;
          }
        }
        kropki_freegame(simgame);
      }
    }
  }
  /* make the move */
  if (bestscore + worstscore < 0) {
    kropki_setpos(game, worstmovex, worstmovey, ai_player);
  } else {
    kropki_setpos(game, bestmovex, bestmovey, ai_player);
  }
}


void ai_play(enum ai_model model, int ai_player, struct kropki_game *game) {
  static int rand_init = 0;
  /* init the random generator, if not inited yet */
  if (rand_init == 0) {
    rand_init = 1;
    srand(time(NULL));
  }

  /* verify first that we can make any move at all */
  if (kropki_findfirstfreefield(game, NULL, NULL) != 0) return;

  /* execute the AI now */
  switch (model) {
    case AI_IDIOT:
      ai_idiot(ai_player, game);
      break;
    case AI_SMART:
      ai_smart(ai_player, game);
      break;
    default:
      puts("UNKNOWN AI MODEL CALLED");
      break;
  }
}
