/*
 * Graphic driver for openKropki, using SDL as backend
 * Copyright (C) 2014, 2015, Mateusz Viste
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdio.h>    /* puts(), printf() */
#include <stdlib.h>   /* malloc(), free() */
#include <SDL2/SDL.h>
#include <SDL2/SDL_ttf.h>
#include <SDL2/SDL2_gfxPrimitives.h>

#include "gz.h"
#include "drv_gra.h" /* include self for control */

#define SCALE 1

struct gra_sprite {
  SDL_Texture *ptr;
  int w;
  int h;
};

static SDL_Window *window = NULL;
static SDL_Renderer *renderer = NULL;


/* loads a gziped bmp image from memory into a SDL surface */
static SDL_Surface *loadgzbmp_surface(unsigned char *memgz, long memgzlen) {
  unsigned char *rawimage;
  long rawimagelen;
  SDL_Surface *res;
  SDL_RWops *rwop;
  if (isGz(memgz, memgzlen) == 0) return(NULL);
  rawimage = ungz(memgz, memgzlen, &rawimagelen);
  rwop = SDL_RWFromMem(rawimage, rawimagelen);
  res = SDL_LoadBMP_RW(rwop, 0);
  SDL_FreeRW(rwop);
  free(rawimage);
  return(res);
}


/* init the video subsystem */
int gra_init(int width, int height, int flags, char *windowtitle, unsigned char *titleicon, long titleicon_len) {
  int sdl_video_flags = SDL_WINDOW_SHOWN | SDL_WINDOW_RESIZABLE;
  SDL_Surface *titleiconsurface;

  if (flags & GRA_FULLSCREEN) sdl_video_flags |= SDL_WINDOW_FULLSCREEN_DESKTOP;
  SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO);
  window = SDL_CreateWindow(windowtitle, SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, width, height, sdl_video_flags);
  renderer = SDL_CreateRenderer(window, -1, 0);
  if (renderer == NULL) {
    SDL_DestroyWindow(window);
    printf("Renderer could not be created! SDL_Error: %s\n", SDL_GetError());
    return(1);
  }

  /* set the window's titlebar icon */
  if ((titleicon != NULL) && (titleicon_len > 0)) {
    titleiconsurface = loadgzbmp_surface(titleicon, titleicon_len);
    if (titleicon != NULL) {
      SDL_SetWindowIcon(window, titleiconsurface);
      SDL_FreeSurface(titleiconsurface);
    }
  }

  /* set the window's minimum size */
  SDL_SetWindowMinimumSize(window, 300, 200);

  /* init the ttf subsystem */
  TTF_Init();

  /* set up alpha blending */
  /* SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_BLEND); */

  return(0);
}


/* close and clean up the graphic subsystem */
void gra_close(void) {
  SDL_DestroyWindow(window);
  TTF_Quit();
  SDL_Quit();
}


void gra_clear(void) {
  SDL_RenderClear(renderer);
}


void gra_switchfullscreen(void) {
  static int fullscreenflag = 0;
  fullscreenflag ^= 1;
  if (fullscreenflag != 0) {
      SDL_SetWindowFullscreen(window, SDL_WINDOW_FULLSCREEN_DESKTOP);
    } else {
      SDL_SetWindowFullscreen(window, 0);
  }
  SDL_Delay(50); /* wait for 50ms - the video thread needs some time to set things up */
}


void gra_drawpartsprite(struct gra_sprite *sprite, int srcx, int srcy, int srcwidth, int srcheight, int dstx, int dsty) {
  static SDL_Rect srcrect;
  static SDL_Rect dstrect;
  if (sprite == NULL) return;
  dstrect.x = dstx * SCALE;
  dstrect.y = dsty * SCALE;
  dstrect.w = srcwidth * SCALE;
  dstrect.h = srcheight * SCALE;
  srcrect.x = srcx;
  srcrect.y = srcy;
  srcrect.w = srcwidth;
  srcrect.h = srcheight;
  SDL_RenderCopy(renderer, sprite->ptr, &srcrect, &dstrect);
}


void gra_drawsprite_alpha(struct gra_sprite *sprite, int x, int y, int alpha) {
  static SDL_Rect rect;
  if (sprite == NULL) return;
  rect.x = x * SCALE;
  alpha = alpha; /* TODO */
  rect.y = y * SCALE;
  rect.w = sprite->w * SCALE;
  rect.h = sprite->h * SCALE;
  /* if (alpha < 255) SDL_SetAlpha(sprite->ptr, SDL_SRCALPHA, alpha); TODO */
  /* SDL_BlitSurface(sprite->ptr, NULL, screen, &rect); */
  SDL_RenderCopy(renderer, sprite->ptr, NULL, &rect);
  /* if (alpha < 255) SDL_SetAlpha(sprite->ptr, SDL_SRCALPHA, 255); TODO */
}


void gra_drawsprite(struct gra_sprite *sprite, int x, int y) {
  gra_drawsprite_alpha(sprite, x, y, 255);
}


void gra_refresh(void) {
  SDL_RenderPresent(renderer);
}


/* loads a gziped bmp image from memory and returns a gra_sprite */
struct gra_sprite *gra_loadgzbmp(unsigned char *memgz, long memgzlen) {
  SDL_Surface *surface;
  struct gra_sprite *res;
  surface = loadgzbmp_surface(memgz, memgzlen);
  if (surface == NULL) return(NULL);
  res = malloc(sizeof(struct gra_sprite));
  res->ptr = SDL_CreateTextureFromSurface(renderer, surface);
  SDL_FreeSurface(surface);
  SDL_QueryTexture(res->ptr, NULL, NULL, &(res->w), &(res->h));
  return(res);
}


void loadSpriteSheet(struct gra_sprite **sprites, int width, int height, int itemcount, void *memptr, int memlen) {
  SDL_Surface *spritesheet, *item;
  SDL_Rect rect;
  int i;
  spritesheet = loadgzbmp_surface(memptr, memlen);
  if (spritesheet == NULL) puts("bmp is NULL!!!");
  item = SDL_CreateRGBSurface(SDL_SWSURFACE, width, height, 32, 0xFF000000L, 0x00FF0000L, 0x0000FF00L, 0x000000FFL);
  if (item == NULL) puts("item is NULL!!!");
  for (i = 0; i < itemcount; i++) {
    rect.x = i * width;
    rect.y = 0;
    rect.w = width;
    rect.h = height;
    SDL_FillRect(item, NULL, 0);
    SDL_BlitSurface(spritesheet, &rect, item, NULL);
    sprites[i] = malloc(sizeof(struct gra_sprite));
    sprites[i]->ptr = SDL_CreateTextureFromSurface(renderer, item);
    if (sprites[i]->ptr == NULL) puts("sprites[i]->ptr is NULL!!!");
    SDL_QueryTexture(sprites[i]->ptr, NULL, NULL, &(sprites[i]->w), &(sprites[i]->h));
  }
  SDL_FreeSurface(item);
  SDL_FreeSurface(spritesheet);
}


int gra_getspritewidth(struct gra_sprite *sprite) {
  if (sprite == NULL) return(0);
  return(sprite->w);
}


int gra_getspriteheight(struct gra_sprite *sprite) {
  if (sprite == NULL) return(0);
  return(sprite->h);
}


int gra_drawrect(int x, int y, int width, int height, int r, int g, int b, int a, int fillflag) {
  SDL_Rect rect;
  int res;
  rect.x = x;
  rect.y = y;
  rect.w = width;
  rect.h = height;
  res = SDL_SetRenderDrawColor(renderer, r, g, b, a);
  if (res != 0) return(-1);
  if (fillflag == 0) {
      res = SDL_RenderDrawRect(renderer, &rect);
    } else {
      res = SDL_RenderFillRect(renderer, &rect);
  }
  if (res != 0) return(-2);
  return(0);
}


void gra_drawline(int x1, int y1, int x2, int y2, int thickness, int r, int g, int b, int a) {
  SDL_SetRenderDrawColor(renderer, r, g, b, a);
  if (x1 < x2) {
      thickLineRGBA(renderer, x1, y1, x2, y2, thickness, r, g, b, a);
    } else {
      thickLineRGBA(renderer, x2, y2, x1, y1, thickness, r, g, b, a);
  }
}


void gra_circle(int x, int y, int rad, int r, int g, int b, int a, int fillflag) {
  if (fillflag != 0) filledCircleRGBA(renderer, x, y, rad, r, g, b, a);
  aacircleRGBA(renderer, x, y, rad, r, g, b, a);
  /* circleRGBA(renderer, x, y, rad, r, g, b, a); */
}


struct gra_sprite *gra_text2sprite(char *message, unsigned char *font_ttf, long font_ttf_len, int fontSize, int r, int g, int b, int a) {
  SDL_Surface *surface;
  SDL_Color txtcolor;
  SDL_RWops *rwop;
  TTF_Font *font = NULL;
  struct gra_sprite *res;

  /* Load the font */
  rwop = SDL_RWFromConstMem(font_ttf, font_ttf_len);
  if (rwop == NULL) {
    puts("OOOPS rwop is NULL!!!!");
    return(NULL);
  }
  font = TTF_OpenFontRW(rwop, 1, fontSize);
  txtcolor.r = r;
  txtcolor.g = g;
  txtcolor.b = b;
  txtcolor.a = 0xff; /* alpha here doesn't work anyway, I set it later on the texture */
  surface = TTF_RenderUTF8_Blended(font, message, txtcolor);
  res = malloc(sizeof(struct gra_sprite));
  res->ptr = SDL_CreateTextureFromSurface(renderer, surface);
  SDL_FreeSurface(surface);
  SDL_QueryTexture(res->ptr, NULL, NULL, &(res->w), &(res->h));
  SDL_SetTextureAlphaMod(res->ptr, a);
  TTF_CloseFont(font);
  /* SDL_FreeRW(rwop); Don't try to free the rwop, it's already free! */
  return(res);
}


void gra_freesprite(struct gra_sprite *spr) {
  if (spr == NULL) return;
  SDL_DestroyTexture(spr->ptr);
  free(spr);
}


void gra_getwinsize(int *w, int *h) {
  SDL_GetWindowSize(window, w, h);
}
